#!/usr/bin/env python2.7
# encoding: utf-8
'''
install_scenario -- install aclib scenarios

install_scenario is a simple script 
that checks what is already installed
and installs new scenario upon request

@author:     Marius Lindauer
        
@copyright:  2013 AClib. All rights reserved.
        
@license:    GPL

@contact:    lindauer@informatik.uni-freiburg.de
'''

import sys
import os
import traceback
import json
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter

from downloader import Downloader
from examiner import Examiner
from update_checker import Updater

__all__ = []
__version__ = 0.2
__date__ = '2013-11-19'
__updated__ = '2014-02-12'

class Installer(object):
    
    def __init__(self, config_file):
        '''
            Constructor
        '''
        try:
            config_ = json.load(open(config_file,"r"))
            self._url = config_["url"]
            self._scenarios = config_["scenarios"]
            self._algorithms = config_["algorithms"]
            self._instances = config_["instances"]
        except:
            sys.stderr.write("[ERROR]: Could not read config file\n")
            traceback.print_exc()
            sys.exit(2)
        
    #===========================================================================
    # def install_scenario(self, scenario_key, scenario_dict=None):
    #     '''
    #         install <scenario>
    #         if <scenario> is a string, the scenario is looked up
    #         if <scenario> is a dictionary, it will be used directly
    #     '''
    #     if scenario_dict is None:
    #         scenario_dict = self._scenarios.get(scenario_key)
    #         if not scenario_dict:
    #             sys.stderr.write("[WARNING]: Have not found scenario: %s\n" %(scenario_key))
    #             return False
    #             
    #     print("Install: %s" %(scenario_key))
    #     print(json.dumps(scenario_dict, indent=2))
    #     
    #     # download instances
    #     instance_set = scenario_dict["instances"]
    #     if not examiner.examine_instances(instance_set, self._instances[instance_set]):
    #         print("Download and Install instance set: %s" %(instance_set))
    #         success = downloader.download_instances(instance_set, self._url, self._instances[instance_set])
    #         if success:
    #             sys.stdout.write("Download Instances successful\n")
    #         else:
    #             sys.stderr.write("[WARNING]: Download Instances failed\n")
    #     
    #     find ASP/ -type f -exec md5sum {} \; | awk '{print $1}' | sort | md5sum
    #===========================================================================
        
    def install_algorithms(self, algorithms):
        '''
            install all <algorithms>
            Args: 
                set of algorithms
        '''
        
        downloader = Downloader()
        examiner = Examiner()
        
        not_installed_algos = []
        
        for algo in algorithms:
            # check algorithms
            if not examiner.examine_algorithm(algo, self._algorithms[algo]["problem_type"]):
                not_installed_algos.append(algo)

        installed_algos = set(algorithms).difference(not_installed_algos)
        
        sys.stdout.write("Already installed algorithms: %s\n" %(",".join(installed_algos)))

        if not not_installed_algos:
            sys.stdout.write("All algorithms are already downloaded and installed.\n")
            return True
        
        sys.stdout.write("To install: %s\n" %(",".join(not_installed_algos)))
        
        return downloader.download_algorithms(not_installed_algos, self._url, self._algorithms)
    
    def install_instances(self, instances):
        '''
            install all <instances>
            Args:
                set of instances
        '''
        
        downloader = Downloader()
        examiner = Examiner()
        
        not_installed_insts = []
        
        for inst in instances:
            # check algorithms
            if not examiner.examine_instances(inst, self._instances[inst]):
                not_installed_insts.append(inst)

        installed_inst = set(instances).difference(not_installed_insts)
        
        sys.stdout.write("Already downloaded instances: %s\n" %(",".join(installed_inst)))
        
        if not not_installed_insts:
            sys.stdout.write("All instance sets are already downloaded and decompressed.\n")
            return True
        
        sys.stdout.write("To install: %s\n" %(",".join(not_installed_insts)))
        
        return downloader.download_instances(not_installed_insts, self._url, self._instances)
    
    def collect_required_ressources(self, scenarios):
        '''
            given a list of ressources, collect all algorithms and instance sets
            Args:
                list of scenarios IDs
            Return:
                list of algorithm IDs
                list of instance IDs
        '''
        
        sys.stdout.write("Installation of %s\n" %(",".join(scenarios)))
        
        algorithms = set()
        instances = set()
        for scenario_id in scenarios:
            scenario = self._scenarios.get(scenario_id)
            algorithms.add(scenario["algorithm"])
            instances.add(scenario["instances"])
        return algorithms, instances
    
    def install_single_scenario(self, scenario):
        '''
            install only one <scenario>
        '''
        print("\nInstall scenarios: %s" %(scenario))
        if not self._scenarios.get(scenario):
            sys.stderr("Scenario not found!")
            return False
        algorithms, instances = self.collect_required_ressources([scenario]) 
        self.install_algorithms(algorithms)
        self.install_instances(instances)
        
    def install_all_with_algorithm(self, algorithm):
        '''
            install all scenarios with <algorithm>
        '''
        print("\nInstall all scenarios with algorithm: %s" %(algorithm))
        selected_scenarios = []
        for id_, scenario in self._scenarios.items():
            if scenario["algorithm"] == algorithm:
                selected_scenarios.append(id_)
        if not selected_scenarios:
            sys.stderr.write("No scenarios found!\n")
            return False
        algorithms, instances = self.collect_required_ressources(selected_scenarios) 
        self.install_algorithms(algorithms)
        self.install_instances(instances)
        
    def install_all_with_problem_type(self, problem_type):
        '''
            install all scenarios with <problem_type>
        '''
        selected_scenarios = []
        print("\nInstall all scenarios with problem type: %s" %(problem_type))
        for id_, scenario in self._scenarios.items():
            instance_set = scenario["instances"]
            if self._instances[instance_set]["problem_type"] == problem_type:
                selected_scenarios.append(id_)
        if not selected_scenarios:
            sys.stderr.write("No scenarios found!\n")
            return False
        algorithms, instances = self.collect_required_ressources(selected_scenarios) 
        self.install_algorithms(algorithms)
        self.install_instances(instances)
                
    def print_scenarios(self):
        '''
            print all scenarios found in config file
        '''
        print("\nScenarios:")
        print("\n".join(list(sorted(self._scenarios.keys()))))
        
    def print_algorithms(self):
        '''
            print all algorithm names found in config file
        '''
        algorithms = set()
        for scenario in self._scenarios.values():
            algorithm = scenario["algorithm"]
            algorithms.add(algorithm)
            
        print("\nAlgorithms:")
        print("\n".join(sorted(algorithms)))
        
    def print_problem_types(self):
        '''
            print all possible problem types
        '''
        problem_types = set()
        for scenario in self._scenarios.values():
            instance_set = scenario["instances"]
            problem = self._instances[instance_set]["problem_type"]
            problem_types.add(problem)
        
        print("\nProblem Types:")
        print("\n".join(sorted(problem_types)))
    

def main(argv=None): # IGNORE:C0111
    '''Command line options.'''
    
    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)
        
    if len(sys.argv) == 1: # no arguments
        sys.argv.append("--help")

    #program_name = os.path.basename(sys.argv[0])
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)
    program_shortdesc = __import__('__main__').__doc__.split("\n")[1]
    program_license = '''%s

  Created by Marius Lindauer on %s.
  Copyright 2013 - AClib. All rights reserved.
  
  Licensed under the GPLv2
  http://www.gnu.org/licenses/gpl-2.0.html
  
  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
''' % (program_shortdesc, str(__date__))

    default_config_file = os.path.join(os.path.split(sys.argv[0])[0], "data", "config.json")
    try:
        # Setup argument parser
        parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
        install_group = parser.add_argument_group("Install")
        install_group.add_argument("-s", "--scenario", dest="scenario", default=None, help="install scenario")
        install_group.add_argument("-a", "--algorithm", dest="algorithm", default=None, help="install scenarios with <algorithm>", metavar="algorithm")
        install_group.add_argument("-p", "--problem", dest="problem", default=None, help="install scenarios with <problem_type>", metavar="problem_type")
        
        show_group = parser.add_argument_group("Show")
        show_group.add_argument("--show_scenarios", dest="show_scenarios", action="store_true", default=False, help="show all possible scenarios")
        show_group.add_argument("--show_algorithms", dest="show_algorithms", action="store_true", default=False, help="show all possible algorithms")
        show_group.add_argument("--show_problems", dest="show_problems", action="store_true", default=False, help="show all possible problem types")
        
        misc_group = parser.add_argument_group("Misc")
        misc_group.add_argument("-C", "--config", dest="config_file", default=default_config_file, help="path to config file in json")
        parser.add_argument('-V', '--version', action='version', version=program_version_message)
        parser.add_argument('-U', '--update', dest="update", action="store_true", default=False, help="check whether update of your aclib version is available")
        
        # Process arguments
        args = parser.parse_args()
        
        if not os.path.isfile(args.config_file):
            sys.stderr.write("[ERROR]: Have not found: %s\n" %(args.config_file))
            sys.exit(1)
            
        if args.update:
            updater = Updater(args.config_file)
            updater.check_update()
        
        installer = Installer(args.config_file)
        
        if args.show_scenarios:
            installer.print_scenarios()
        if args.show_algorithms:
            installer.print_algorithms()
        if args.show_problems:
            installer.print_problem_types()
        
        if args.scenario:
            installer.install_single_scenario(args.scenario)
        if args.algorithm:
            installer.install_all_with_algorithm(args.algorithm)
        if args.problem:
            installer.install_all_with_problem_type(args.problem)
        
        
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 0

if __name__ == "__main__":
    sys.exit(main())
