#!/usr/bin/env python2.7
# encoding: utf-8
'''
run_scenario -- run a given AClib scenario 

run_scenario is a script to create a working
directory and run a specific AClib scenario.
It is assumed that all resources required by
the scenario have already been installed
using install_scenario.py

@author:     Chris Fawcett, Marius Lindauer
        
@copyright:  2013 AClib. All rights reserved.
        
@license:    GPL

@contact:    fawcettc@cs.ubc.ca
'''

import sys
import os
import traceback
import json
import subprocess
import time

from cmd_parser.run_parser import CMD_Parser

__all__ = []
__version__ = 0.1
__date__ = '2013-12-01'
__updated__ = '2013-12-31'

class ScenarioRunner(object):

    def __init__(self, config_file, scenario_id, aclib_root, working_directory):
        '''
            Constructor
        '''
        config_ = json.load(open(config_file, "r"))
        self.aclib_scenarios = config_["scenarios"]
        self.aclib_algorithms = config_["algorithms"]
        self.aclib_instances = config_["instances"]

        self.aclib_root_path = os.path.abspath(aclib_root)
        self.working_directory_path = os.path.abspath(working_directory)
        self.scenario_id = scenario_id

        self.configurator_type = "dummy"
        self.problem_type = "dummy"
        self.seed = 1
        self.subprocesses = []
        
        self.DELAY2KILL = 2

    def prepare_working_directory(self):
        if not os.path.exists(self.working_directory_path):
            os.makedirs(self.working_directory_path)
        else:
            sys.stderr.write("[WARNING] Working directory '%s' already exists. Going forward anyway...\n" % (self.working_directory_path))

        os.chdir(self.working_directory_path)

        scenarios_path = os.path.join(self.aclib_root_path, "scenarios")
        algorithms_path = os.path.join(self.aclib_root_path, "target_algorithms")
        instances_path = os.path.join(self.aclib_root_path, "instances")

        configurators_path = os.path.join(self.aclib_root_path, "configurators")

        if not os.path.exists("./scenarios"):
            os.symlink(scenarios_path, "./scenarios")
        else:
            sys.stderr.write("[WARNING] ./scenarios existed in working directory. Assuming correct symlink...\n")

        if not os.path.exists("./target_algorithms"):
            os.symlink(algorithms_path, "./target_algorithms")
        else:
            sys.stderr.write("[WARNING] ./target_algorithms existed in working directory. Assuming correct symlink...\n")

        if not os.path.exists("./instances"):
            os.symlink(instances_path, "./instances")
        else:
            sys.stderr.write("[WARNING] ./instances existed in working directory. Assuming correct symlink...\n")

        if not os.path.exists("./configurators"):
            os.symlink(configurators_path, "./configurators")
        else:
            sys.stderr.write("[WARNING] ./configurators existed in working directory. Assuming correct symlink...\n")

    def check_resource_requirements(self):
        scenario_dict = self.aclib_scenarios.get(self.scenario_id)

        if not scenario_dict:
            sys.stderr.write("[ERROR] Scenario id %s could not be found in the configuration file!\n" % (self.scenario_id))
            sys.exit(1)

        algorithm_dict = self.aclib_algorithms.get(scenario_dict.get("algorithm"))
        instances_dict = self.aclib_instances.get(scenario_dict.get("instances"))

        if not algorithm_dict:
            sys.stderr.write("[ERROR] Scenario id %s requires algorithm %s which was not found in the configuration file!\n" % (self.scenario_id, scenario_dict.get("algorithm")))
            sys.exit(1)

        if not instances_dict:
            sys.stderr.write("[ERROR] Scenario id %s requires instance set %s which was not found in the configuration file!\n" % (self.scenario_id, scenario_dict.get("instances")))
            sys.exit(1)

        supported_configurators = scenario_dict.get("configurators")

        if self.configurator_type not in supported_configurators:
            sys.stderr.write("[ERROR] Scenario id %s does not support the requested configurator %s!\n" % (self.scenario_id, self.configurator_type))
            sys.exit(1)

        self.problem_type = instances_dict.get("problem_type")
 
        print(json.dumps(scenario_dict,indent=2))
        print("problem type = %s" % (self.problem_type))

    def build_command_line(self):
        self.scenario_command_line = ["echo", "dummy configurator"]

    def dump_command_line(self):
        return " ".join(self.scenario_command_line)

    def prepare(self, seed):
        self.seed = seed
        self.prepare_working_directory()
        self.check_resource_requirements()
        self.build_command_line()
        self.build_validate_line()

    def cleanup(self):
        if (len(self.subprocesses) > 0):
            print("killing the scenario!")
            for sub in self.subprocesses:
                #sub.terminate()
                subprocess.Popen(["pkill","-TERM", "-P",str(sub.pid)])
                print("Wait %d seconds ..." %(self.DELAY2KILL))
                time.sleep(self.DELAY2KILL)
                if sub.returncode is None: # still running
                    sub.kill()
            print("done... If anything in the subprocess tree fork'd a new process group, we may not have caught everything...")

    def run_scenario(self):
        cmd = " ".join(self.scenario_command_line)
        print("Running %s" % (cmd))
        io = subprocess.Popen(cmd, shell=True, preexec_fn=os.setpgrp)
        self.subprocesses.append(io)
        print("Waiting on subprocess completion...")
        io.wait()
        self.subprocesses.remove(io)
        print("Configuration finished.")

    def run_validate(self):
        cmd = " ".join(self.validate_line)
        print("Running %s" % (cmd))
        io = subprocess.Popen(cmd, shell=True, preexec_fn=os.setpgrp)
        self.subprocesses.append(io)
        print("Waiting on subprocess completion...")
        io.wait()
        self.subprocesses.remove(io)
        cmd = " ".join(self.validate_def_line)
        print("Running %s" % (cmd))
        io = subprocess.Popen(cmd, shell=True, preexec_fn=os.setpgrp)
        self.subprocesses.append(io)
        print("Waiting on subprocess completion...")
        io.wait()
        self.subprocesses.remove(io)
        print("Validation finished.")

class ParamILSRunner(ScenarioRunner):

    def __init__(self, config_file, scenario_id, aclib_root, working_directory):
        '''
            Constructor
        '''
        ScenarioRunner.__init__(self, config_file, scenario_id, aclib_root, working_directory)
        self.configurator_type = "ParamILS"

    def build_command_line(self):
        self.scenario_command_line = ["ruby", "./configurators/paramils/param_ils_2_3_run.rb", "-scenariofile", "./scenarios/%s/%s/scenario.txt" % (self.problem_type, self.scenario_id), "-numRun", str(self.seed), "-validN", "0"]

    def build_validate_line(self):
        self.validate_line = ["./configurators/smac/smac-validate", "--scenarioFile", "./scenarios/%s/%s/scenario.txt" % (self.problem_type, self.scenario_id), "--numRun", str(self.seed), "--trajectoryFile", "./paramils_out/*traj_%d.csv" %(self.seed), "--outputDirectory", "./validate-test", "--validateOnlyLastIncumbent", "true", "--numValidationRuns", "1", "--useScenarioOutDir", "true"]
        self.validate_def_line = ["./configurators/smac/smac-validate", "--scenarioFile", "./scenarios/%s/%s/scenario.txt" % (self.problem_type, self.scenario_id), "--numRun", str(self.seed), "--outputDirectory", "./validate-def", "--includeDefaultAsFirstRandom", "true", "--random-configurations", "1", "--numValidationRuns", "1", "--useScenarioOutDir", "true"]

class SMACRunner(ScenarioRunner):

    def __init__(self, config_file, scenario_id, aclib_root, working_directory):
        '''
            Constructor
        '''
        ScenarioRunner.__init__(self, config_file, scenario_id, aclib_root, working_directory)
        self.configurator_type = "SMAC"

    def build_command_line(self):
        self.scenario_command_line = ["./configurators/smac/smac", "--validation", "false", "--scenarioFile", "./scenarios/%s/%s/scenario.txt" % (self.problem_type, self.scenario_id), "--seed", str(self.seed), "--log-all-call-strings", "true", "--rungroup", "aclib"]

    def build_validate_line(self):
        self.validate_line = ["./configurators/smac/smac-validate", "--scenarioFile", "./scenarios/%s/%s/scenario.txt" % (self.problem_type, self.scenario_id), "--numRun", str(self.seed), "--trajectoryFile", "./smac-output/aclib/traj-run-%d.csv" %(self.seed), "--outputDirectory", "./validate-inc", "--validateOnlyLastIncumbent", "true", "--numValidationRuns", "1", "--useScenarioOutDir", "true"]
        self.validate_def_line = ["./configurators/smac/smac-validate", "--scenarioFile", "./scenarios/%s/%s/scenario.txt" % (self.problem_type, self.scenario_id), "--numRun", str(self.seed), "--outputDirectory", "./validate-def", "--includeDefaultAsFirstRandom", "true", "--random-configurations", "1", "--numValidationRuns", "1", "--useScenarioOutDir", "true"]

class IRaceRunner(ScenarioRunner):

    def __init__(self, config_file, scenario_id, aclib_root, working_directory):
        '''
            Constructor
        '''
        ScenarioRunner.__init__(self, config_file, scenario_id, aclib_root, working_directory)
        self.configurator_type = "irace"

    def build_command_line(self):
        self.scenario_command_line = ["./configurators/irace/inst/bin/irace", "--config-file ./scenarios/%s/%s/scenario.txt" % (self.problem_type, self.scenario_id), "--seed", str(self.seed), "--exec-dir ."]

    def build_validate_line(self):
        self.validate_line = ["NOT YET IMPLEMENTED"]

def print_scenarios(config_file):
    '''
        print all scenarios found in config file
    '''
    try:
        config = json.load(open(config_file, "r"))
        scenarios = config["scenarios"]

        print("\nScenarios:")
        print("\n".join(list(sorted(scenarios.keys()))))
    except:
        sys.stderr.write("[ERROR] Could not read config file %s\n" % (config_file))
        traceback.print_exc()
        sys.exit(2)


def main(argv=None): # IGNORE:C0111
    '''Command line options.'''
    
    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    #program_name = os.path.basename(sys.argv[0])
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = "%%(prog)s %s (%s)" % (program_version, program_build_date)
    program_shortdesc = __import__("__main__").__doc__.split("\n")[1]
    program_license = '''%s

  Created by Chris Fawcett on %s.
  Copyright 2013 - AClib. All rights reserved.
  
  Licensed under the GPLv2
  http://www.gnu.org/licenses/gpl-2.0.html
  
  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
''' % (program_shortdesc, str(__date__))

    aclib_root = os.path.split(os.path.split(__file__)[0])[0]
    runner = None
    try:
        # Setup argument parser
        parser = CMD_Parser(program_license, program_version_message, aclib_root)
        # Process arguments
        args = parser.parse()
        
        if not os.path.isfile(args.config_file):
            sys.stderr.write("[ERROR]: AClib configuration file not found: %s\n" %(args.config_file))
            sys.exit(1)
        
        current_wd = os.getcwd()

        if args.show_scenarios:
            print_scenarios(args.config_file)
            return 0

        if (args.configurator == "ParamILS"):
                runner = ParamILSRunner(args.config_file, args.scenario, aclib_root, args.working_directory)
        elif (args.configurator == "SMAC"):
                runner = SMACRunner(args.config_file, args.scenario, aclib_root, args.working_directory)
        elif (args.configurator == "irace"):
                runner = IRaceRunner(args.config_file, args.scenario, aclib_root, args.working_directory)
            # Additional configurators here...
        else:
                sys.stderr.write("[ERROR]: Unrecognized configurator %s, dying.\n" % (args.configurator))
                return 1
    
        runner.prepare(args.seed)
        
        if args.print_cmd_only: 
            print "Configurator command-line (from working directory):"
            print runner.dump_command_line()
        else:
            runner.run_scenario()
            if args.validate:
                runner.run_validate()

    except (KeyboardInterrupt, SystemExit):
        ### handle keyboard interrupt ###
        if runner:
            runner.cleanup()
        return 1

if __name__ == "__main__":
    sys.exit(main())
