#!/usr/bin/env python2.7
# encoding: utf-8
'''
test_scenarios -- test all AClib scenarios 

test_scenarios is a script to create a working
directory and verify all installed scenarios with the help of SMAC.
It is assumed that all resources required by
the scenario have already been installed
using install_scenario.py

@author:     Marius Lindauer; Chris Fawcett
        
@copyright:  2013 AClib. All rights reserved.
        
@license:    GPL

@contact:    fawcettc@cs.ubc.ca
'''

import sys
import os
import traceback
import json
from subprocess import Popen, PIPE
import time

from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter

__all__ = []
__version__ = 0.1
__date__ = '2014-02-14'
__updated__ = '2014-02-14'

class ScenearioChecker(object):

    def __init__(self, config_file, aclib_root, working_directory, check_algos):
        '''
            Constructor
        '''
        config_ = json.load(open(config_file, "r"))
        self.aclib_scenarios = config_["scenarios"]
        self.aclib_algorithms = config_["algorithms"]
        self.aclib_instances = config_["instances"]

        self.aclib_root_path = os.path.abspath(aclib_root)
        self.working_directory_path = os.path.abspath(working_directory)

        self.check_algos = check_algos

        self.configurator_type = "dummy"
        self.problem_type = "dummy"
        self.seed = 1
        self.subprocesses = []
        self.current_wd = os.getcwd()
        
        self.DELAY2KILL = 2

    def prepare_working_directory(self):
        if not os.path.exists(self.working_directory_path):
            os.makedirs(self.working_directory_path)
        else:
            sys.stderr.write("[WARNING] Working directory '%s' already exists. Going forward anyway...\n" % (self.working_directory_path))

        os.chdir(self.working_directory_path)

        scenarios_path = os.path.join(self.aclib_root_path, "scenarios")
        algorithms_path = os.path.join(self.aclib_root_path, "target_algorithms")
        instances_path = os.path.join(self.aclib_root_path, "instances")

        configurators_path = os.path.join(self.aclib_root_path, "configurators")

        if not os.path.exists("./scenarios"):
            os.symlink(scenarios_path, "./scenarios")
        else:
            sys.stderr.write("[WARNING] ./scenarios existed in working directory. Assuming correct symlink...\n")

        if not os.path.exists("./target_algorithms"):
            os.symlink(algorithms_path, "./target_algorithms")
        else:
            sys.stderr.write("[WARNING] ./target_algorithms existed in working directory. Assuming correct symlink...\n")

        if not os.path.exists("./instances"):
            os.symlink(instances_path, "./instances")
        else:
            sys.stderr.write("[WARNING] ./instances existed in working directory. Assuming correct symlink...\n")

        if not os.path.exists("./configurators"):
            os.symlink(configurators_path, "./configurators")
        else:
            sys.stderr.write("[WARNING] ./configurators existed in working directory. Assuming correct symlink...\n")

    def check_resource_requirements(self, scenario_id):
        '''
            checks resources and returns False if anything is missing
        '''
        scenario_dict = self.aclib_scenarios.get(scenario_id)

        if not scenario_dict:
            sys.stderr.write("[ERROR] Scenario id %s could not be found in the configuration file!\n" % (scenario_id))
            return False

        algorithm_dict = self.aclib_algorithms.get(scenario_dict.get("algorithm"))
        instances_dict = self.aclib_instances.get(scenario_dict.get("instances"))
        problem_type = instances_dict["problem_type"]

        algo_dir = os.path.join(self.aclib_root_path,"target_algorithms",problem_type,scenario_dict.get("algorithm"))
        if (not algorithm_dict) or not os.path.exists(algo_dir):
            sys.stderr.write("[ERROR] Algorithm not installed (%s: %s)!\n" % (scenario_id, algo_dir))
            return False
        
        inst_dir = os.path.join(self.aclib_root_path,"instances",problem_type,"data",scenario_dict.get("instances"))
        if problem_type != "ml" and (not instances_dict or not os.path.exists(inst_dir)):
            sys.stderr.write("[ERROR] Instances not installed (%s: %s)!\n" % (scenario_id, inst_dir))
            return False

        supported_configurators = scenario_dict.get("configurators")

        if "SMAC" not in supported_configurators:
            sys.stderr.write("[ERROR] Scenario id %s does not support SMAC!  Cannot check further\n" % (scenario_id))
            return False

        #self.problem_type = instances_dict.get("problem_type")
 
        #print(json.dumps(scenario_dict,indent=2))
        #print("problem type = %s" % (self.problem_type))
        return True

    def check_scenario(self, scenario_id):
        '''
            check senerio with the help of verify-scenario
        '''
        scenario_dict = self.aclib_scenarios.get(scenario_id)
        instances_dict = self.aclib_instances.get(scenario_dict.get("instances"))
        problem_type = instances_dict.get("problem_type")
        
        cmd = ["./configurators/smac-v2.06.01-master-622/verify-scenario", "--scenarios", "./scenarios/%s/%s/scenario.txt" % (problem_type, scenario_id)]
        
        if problem_type == "ml":
            cmd.extend(["--verify-instances", "false"])
        
        print(" ".join(cmd))
        p = Popen(cmd, stdout=PIPE)
        self.subprocesses.append(p)
        (out_, err_) = p.communicate()
        self.subprocesses.remove(p)
        
        return_value = True
        for line in out_.split("\n"):
            print(line)
            if line.startswith("[ERROR]"):
                return_value = False
        return return_value

        
    def check_algo(self, scenario_id):
        '''
            check algo with the help of algo-test
        '''
        scenario_dict = self.aclib_scenarios.get(scenario_id)
        instances_dict = self.aclib_instances.get(scenario_dict.get("instances"))
        problem_type = instances_dict.get("problem_type")
        
        cmd = ["./configurators/smac-v2.06.01-master-622/algo-test", "--scenarioFile", "./scenarios/%s/%s/scenario.txt" % (problem_type, scenario_id)]
        print(" ".join(cmd))
        p = Popen(cmd, stdout=PIPE)
        self.subprocesses.append(p)
        (out_, err_) = p.communicate()
        self.subprocesses.remove(p)
        
        return_value = True
        for line in out_.split("\n"):
            print(line)
            if line.startswith("[ERROR]"):
                return_value = False
        return return_value
        
    def build_command_line(self):
        self.scenario_command_line = ["echo", "dummy configurator"]

    def dump_command_line(self):
        return " ".join(self.scenario_command_line)

    def prepare(self):
        self.prepare_working_directory()
        self.build_command_line()

    def cleanup(self):
        os.chdir(self.current_wd)
        if (len(self.subprocesses) > 0):
            print("killing the scenario!")
            for sub in self.subprocesses:
                sub.terminate()
                print("Wait %d seconds ..." %(self.DELAY2KILL))
                time.sleep(self.DELAY2KILL)
                if sub.returncode is None: # still running
                    sub.kill()
                #TODO: Kill entire process tree
            print("done... If anything in the subprocess tree fork'd a new process group, we may not have caught everything...")

    def check_scenarios(self, scenario=None):
        '''
            checks every scenarios:
            1. installed? (check resources)
            2. running scenario checker of SMAC
            3. running algorithm checker of SMAC
        '''
        scenario_status = {}
        for scenario_id in self.aclib_scenarios.keys():
            if scenario and scenario_id != scenario:
                continue
            print("check %s ..." %(scenario_id))
            if not self.check_resource_requirements(scenario_id): 
                scenario_status[scenario_id] = "Resources missing (or SMAC not supported)"
                continue
            if not self.check_scenario(scenario_id) : 
                scenario_status[scenario_id] = "Scenario failed"
                continue
            if self.check_algos and not self.check_algo(scenario_id):
                scenario_status[scenario_id] = "Target algorithm failed"
                continue
            scenario_status[scenario_id] = "ok"
        for scenario in sorted(list(scenario_status.keys())):
            print("%s\t ... %s" %(scenario, scenario_status[scenario]))

def print_scenarios(config_file):
    '''
        print all scenarios found in config file
    '''
    try:
        config = json.load(open(config_file, "r"))
        scenarios = config["scenarios"]

        print("\nScenarios:")
        print("\n".join(list(scenarios.keys())))
    except:
        sys.stderr.write("[ERROR] Could not read config file %s\n" % (config_file))
        traceback.print_exc()
        sys.exit(2)


def main(argv=None): # IGNORE:C0111
    '''Command line options.'''
    
    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    #program_name = os.path.basename(sys.argv[0])
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = "%%(prog)s %s (%s)" % (program_version, program_build_date)
    program_shortdesc = __import__("__main__").__doc__.split("\n")[1]
    program_license = '''%s

  Created by Marius Lindauer, Chris Fawcett on %s.
  Copyright 2013 - AClib. All rights reserved.
  
  Licensed under the GPLv2
  http://www.gnu.org/licenses/gpl-2.0.html
  
  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
''' % (program_shortdesc, str(__date__))

    aclib_root = os.path.split(os.path.split(__file__)[0])[0]
    default_config_file = os.path.join(aclib_root, "src", "data", "config.json")

    try:
        # Setup argument parser
        parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
        run_group = parser.add_argument_group("Run")
        run_group.add_argument("-d", "--working_directory", dest="working_directory", default=".", help="working directory")
        run_group.add_argument("-a", "--check_algorithms", dest="check_algos", default=False, action="store_true",  help="make test call of algorithms (attention: needs a lot of time)")
        run_group.add_argument("-s", "--scenario", dest="scenario", default=None, help="check only the given scenario")

        misc_group = parser.add_argument_group("Misc")
        misc_group.add_argument("-c", "--config", dest="config_file", default=default_config_file, help="path to the AClib configuration file")
        parser.add_argument('-V', '--version', action='version', version=program_version_message)
        
        # Process arguments
        args = parser.parse_args()
        
        if not os.path.isfile(args.config_file):
            sys.stderr.write("[ERROR]: AClib configuration file not found: %s\n" %(args.config_file))
            sys.exit(1)
        
        runner = ScenearioChecker(args.config_file, aclib_root, args.working_directory, args.check_algos)
        runner.current_wd = os.getcwd()
        runner.prepare()
        
        runner.check_scenarios(args.scenario)

    except (KeyboardInterrupt, SystemExit):
        ### handle keyboard interrupt ###
        runner.cleanup()
        return 1

if __name__ == "__main__":
    sys.exit(main())
